{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE ForeignFunctionInterface #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE DataKinds #-}
module IGraph.Algorithms.Community
    ( findCommunity
    , CommunityMethod(..)
    , leadingEigenvector
    , spinglass
    , leiden
    , infomap
    , modularity
    ) where

import           Data.Foldable             (foldl')
import           Data.List.Ordered         (nubSortBy)
import           Data.Ord                  (comparing)
import           System.IO.Unsafe          (unsafePerformIO)
import           Data.Serialize            (Serialize)
import qualified Data.IntMap.Strict        as IM

import           Foreign
import           Foreign.C.Types

import           IGraph
import           IGraph.Random
import IGraph.Internal.C2HS
{#import IGraph.Internal #}
{#import IGraph.Internal.Constants #}

#include "haskell_igraph.h"

-- | Detecting community structure.
findCommunity :: (Serialize v, Serialize e)
              => Graph 'U v e
              -> Maybe (Node -> v -> Double)  -- ^ Function to assign node weights
              -> Maybe (e -> Double)  -- ^ Function to assign edge weights
              -> CommunityMethod  -- ^ Community finding algorithms
              -> Gen
              -> IO [[Int]]
findCommunity gr getNodeW getEdgeW method _ = allocaVectorInt $ \result ->
    withListMaybe ew $ \ew' -> do
        case method of
            LeadingEigenvector n -> allocaArpackOpt $ \arpack ->
                igraphCommunityLeadingEigenvector (_graph gr) ew' nullPtr result
                                                  n arpack nullPtr False
                                                  nullPtr nullPtr nullPtr
                                                  nullFunPtr nullPtr
            Spinglass{..} -> igraphCommunitySpinglass (_graph gr) ew' nullPtr nullPtr result
                                     nullPtr _nSpins False _startTemp
                                     _stopTemp _coolFact
                                     IgraphSpincommUpdateConfig _gamma
                                     IgraphSpincommImpOrig 1.0
            Leiden{..} -> do
                _ <- withListMaybe nw $ \nw' -> igraphCommunityLeiden
                    (_graph gr) ew' nw' _resolution _beta False _lIter result nullPtr
                return ()
            Infomap{..} -> do
                _ <- withListMaybe nw $ \nw' -> alloca $ \codeLenPtr ->
                       igraphCommunityInfomap (_graph gr) ew' nw' _iIter result codeLenPtr
                return ()

        createClusters <$> toIntList result
  where
    -- Processes the membership result, turning that into a list of clusters. What we get from
    -- the C library is something like this [0,0,0,1,1], which length is the number of vertexes
    -- in the graph, and each /positional/ element describes which cluster the given element belongs
    -- to. In the example, nodes 0,1 and 2 belongs to the cluster \"0\", and nodes 3 and 4 belogs to
    -- the cluster \"1\".
    createClusters :: [Int] -> [[Int]]
    createClusters = map reverse . IM.elems . snd . foldl' processMembership (0, mempty)

    processMembership :: (Int, IM.IntMap [Int]) -> Int -> (Int, IM.IntMap [Int])
    processMembership (!currentNodeId, !acc) currentClusterId =
      (currentNodeId + 1, IM.alter (\case Nothing       -> Just $! [currentNodeId]
                                          Just siblings -> Just $! (currentNodeId : siblings)
                                   ) currentClusterId acc)

    ew = case getEdgeW of
        Nothing -> Nothing
        Just f -> Just $ map (f . snd) $ labEdges gr
    nw = case getNodeW of
        Nothing -> Nothing
        Just f -> Just $ map (uncurry f) $ labNodes gr

data CommunityMethod =
      LeadingEigenvector
        { _nIter     :: Int  -- ^ number of iterations, default is 10000
        }
    | Spinglass
        { _nSpins    :: Int  -- ^ number of spins, default is 25
        , _startTemp :: Double  -- ^ the temperature at the start
        , _stopTemp  :: Double  -- ^ the algorithm stops at this temperature
        , _coolFact  :: Double  -- ^ the cooling factor for the simulated annealing
        , _gamma     :: Double  -- ^ the gamma parameter of the algorithm.
        }
    | Leiden
        { _resolution :: Double
        , _beta :: Double
        , _lIter :: Int
        }
    | Infomap { _iIter :: Int  -- ^ number of iterations
               }

-- | Default parameters for the leading eigenvector algorithm.
leadingEigenvector :: CommunityMethod
leadingEigenvector = LeadingEigenvector 10000

-- | Default parameters for the spin-glass algorithm.
spinglass :: CommunityMethod
spinglass = Spinglass
    { _nSpins = 25
    , _startTemp = 1.0
    , _stopTemp = 0.01
    , _coolFact = 0.99
    , _gamma = 1.0 }

-- | Default parameters for the leiden algorithm.
-- 1 / 2m sum_ij (A_ij - gamma n_i n_j)d(s_i, s_j), where
-- m is the total edge weight,
-- A_ij is the weight of edge (i, j),
-- gamma is the so-called resolution parameter,
-- n_i is the node weight of node i,
-- s_i is the cluster of node i and
-- d(x, y) = 1 if and only if x = y and 0 otherwise.
-- By setting n_i = k_i, the degree of node i, and dividing gamma by 2m,
-- you effectively obtain an expression for modularity.
-- Hence, the standard modularity will be optimized when you supply the degrees
-- as node_weights and by supplying as a resolution parameter 1.0/(2*m), with m the number of edges.
--
-- RBConfigurationVertexPartition: supplying the degrees as node weights, and
-- a resolution parameter 1.0/(2*m), with m the number of edges.
-- CPM: 
leiden :: CommunityMethod
leiden = Leiden
    { _resolution = 0.001
    , _beta = 0.001
    , _lIter = 10
    }

infomap :: CommunityMethod
infomap = Infomap 10

{#fun igraph_community_spinglass as ^
    { `IGraph'                     -- graph
    , castPtr `Ptr Vector'         -- weights
    , id `Ptr CDouble'             -- modularity
    , id `Ptr CDouble'             -- temperature
    , castPtr `Ptr VectorInt'      -- membership
    , castPtr `Ptr VectorInt'      -- csize
    , `Int'                        -- spins
    , `Bool'                       -- parupdate
    , `Double'                     -- starttemp
    , `Double'                     -- stoptemp
    , `Double'                     -- coolfact
    , `SpincommUpdate'             -- update_rule
    , `Double'                     -- gamma
    , `SpinglassImplementation'    -- implementation
    , `Double'                     -- gamma_minus
    } -> `CInt' void- #}

{#fun igraph_community_leading_eigenvector as ^
    { `IGraph'                 -- graph
    , castPtr `Ptr Vector'     -- weights
    , castPtr `Ptr MatrixInt'  -- merges
    , castPtr `Ptr VectorInt'  -- membership
    , `Int'                    -- steps
    , castPtr `Ptr ArpackOpt'  -- options
    , id `Ptr CDouble'         -- modularity
    , `Bool'                   -- start
    , castPtr `Ptr Vector'     -- eigenvalues
    , castPtr `Ptr VectorList' -- eigenvectors
    , castPtr `Ptr Vector'     -- histroy
    , castFunPtr `T'           -- callback
    , id `Ptr ()'              -- callback_extra
    } -> `CInt' void- #}

{#fun igraph_community_leiden as ^
    { `IGraph'                      -- graph
    , castPtr `Ptr Vector'          -- edge weights
    , castPtr `Ptr Vector'          -- node weights
    , `Double'                      -- resolution parameter
    , `Double'                      -- beta
    , `Bool'                        -- start
    , `Int'                         -- n_interations
    , castPtr `Ptr VectorInt'       -- membership
    , alloca- `Int' peekIntConv*    -- nb_clusters
    , id `Ptr CDouble'              -- quality
    } -> `CInt' void- #}

{#fun igraph_community_infomap as ^
    { `IGraph'                 -- graph
    , castPtr `Ptr Vector'     -- e_weights
    , castPtr `Ptr Vector'     -- v_weights
    , `Int'                    -- nb_trials
    , castPtr `Ptr VectorInt'  -- membership
    , id `Ptr CDouble'         -- codelength
    } -> `CInt' void- #}


type T = FunPtr ( Ptr ()
                -> Int
                -> CDouble
                -> Ptr ()
                -> FunPtr (Ptr CDouble -> Ptr CDouble -> CInt -> Ptr () -> IO CInt)
                -> Ptr ()
                -> Ptr ()
                -> IO CInt)

-- | Calculate the modularity of a graph with respect to some vertex types.
modularity :: Serialize e
           => Graph d v e
           -> Maybe (e -> Double)  -- ^ Function to assign edge weights
           -> [[Int]]   -- ^ Communities.
           -> Double
modularity gr getEdgeW clusters
    | length nds /= length (concat clusters) = error "Duplicated nodes"
    | nds /= nodes gr = error "Some nodes were not given community assignments"
    | otherwise = unsafePerformIO $ withList membership $ \membership' ->
        withListMaybe ws $ \weightsPtr ->
          igraphModularity (_graph gr) membership' weightsPtr 1.0 False
  where
    (membership, nds) = unzip $ nubSortBy (comparing snd) $ concat $
        zipWith f [0 :: Int ..] clusters
      where
        f i xs = zip (repeat i) xs
    ws = case getEdgeW of
        Nothing -> Nothing
        Just f -> Just $ map (f . snd) $ labEdges gr
-- FIXME(adinapoli) We are defaulting to undirected graphs, and ignoring
-- the 'directed' parameter. Also defaulting to '1' for the resolution
-- parameter.
{#fun igraph_modularity as ^
    { `IGraph'                        -- graph
    , castPtr `Ptr Vector'            -- membership
    , castPtr `Ptr Vector'            -- weight
    , `CDouble'                       -- resolution
    , `Bool'                          -- directed
    , alloca- `Double' peekFloatConv* -- modularity
    } -> `CInt' void- #}

